from dataclasses import dataclass
from pathlib import Path
from typing import Optional


@dataclass
class OutputOptions:
    output_file_path: Optional[Path] = None
    console: bool = False
    scale: bool = False
    log_reading: bool = False
    save_reading: bool = True
    omit_zero_readings: bool = True
    omit_none_readings: bool = True
    add_timestamp_to_reading: bool = False
    add_timestamp_to_filename: bool = True



import logging
from contextlib import contextmanager

import sunspec2.file.client as file_client
import sunspec2.modbus.client as client

from pysunspec_read.connect_options import ConnectOptions

logger = logging.getLogger(__name__)


@contextmanager
def read(connect_options: ConnectOptions):
    device = connect(connect_options)
    logger.info("Reading from device")
    device.scan()
    try:
        yield device
    finally:
        if connect_options.file is None:
            logger.info("Closing device connection")
            device.close()


def connect_file(file):
    logger.info("Connecting to file as input, name=%s, address=%s", file.filename, file.addr)
    return file_client.FileClientDevice(filename=file.filename)


def connect_rtu(rtu):
    logger.info("Connecting to device using device type=rtu, name=%s, baudrate=%s, parity=%s, timeout=%s",
                rtu.name,
                rtu.baudrate,
                rtu.parity,
                rtu.timeout)
    return client.SunSpecModbusClientDeviceRTU(slave_id=rtu.slave_id,
                                               name=rtu.name,
                                               baudrate=rtu.baudrate,
                                               parity=rtu.parity,
                                               timeout=rtu.timeout,
                                               ctx=rtu.ctx,
                                               max_count=rtu.max_count)


def connect_tcp(tcp):
    logger.info("Connecting to device using device type=tcp, ip=%s, port=%s, slave_id=%s, timeout=%s",
                tcp.ip_address,
                tcp.ip_port,
                tcp.slave_id,
                tcp.timeout)
    return client.SunSpecModbusClientDeviceTCP(slave_id=tcp.slave_id,
                                               ipaddr=tcp.ip_address,
                                               ipport=tcp.ip_port,
                                               timeout=tcp.timeout,
                                               ctx=tcp.ctx,
                                               test=tcp.test,
                                               max_count=tcp.max_count)


def connect(connect_options: ConnectOptions):
    if connect_options.tcp is not None:
        return connect_tcp(connect_options.tcp)
    elif connect_options.rtu is not None:
        return connect_rtu(connect_options.rtu)
    elif connect_options.file is not None:
        return connect_file(connect_options.file)
    else:
        raise NameError("No connect option supplied")



import json
import logging
import os
import shutil
import tempfile
from pathlib import Path

log = logging.getLogger(__name__)


def assert_json_files_equals(self, actual_json_path, expected_json_path_relative_to_test):
    expected_path = Path(get_test_folder(), expected_json_path_relative_to_test)
    with open(expected_path, mode='r', newline=None) as expected, open(actual_json_path, 'r') as actual:
        expected_json = expected.read()
        actual_json = actual.read()
        print(actual_json)
        self.assertEqual(expected_json, actual_json)


def assert_json_equals(self, json_str, expected_json_path_relative_to_test):
    expected_path = Path(get_test_folder(), expected_json_path_relative_to_test)
    with open(expected_path, mode='r', newline=None) as expected:
        expected_json = expected.read()
        print(json_str)
        self.assertEqual(expected_json, json_str)


def load_resource_as_string(resource_path_string_relative_to_test):
    resource_path = get_resource_path(resource_path_string_relative_to_test)
    return load_path_as_string(resource_path)


def load_path_as_string(path):
    with open(path, mode='r', newline=None) as expected:
        content = expected.read()
    return content


def load_resource_as_json(json_path_relative_to_test):
    resource_path = get_resource_path(json_path_relative_to_test)
    return load_path_as_json(resource_path)


def load_path_as_json(path):
    content = load_path_as_string(path)
    return json.loads(content)


def get_resource_path(path_relative_to_test_folder: str):
    return Path(get_test_folder(), path_relative_to_test_folder)


def get_test_folder():
    return get_folder()


def get_folder(name: str = "test"):
    nested = 1
    current_folder = Path(".").absolute()
    while current_folder.name != name:
        print("looking for {name} folder, currently at {current}".format(name=name,
                                                                         current=current_folder.name + " " + str(
                                                                             current_folder)))
        current_folder = current_folder.parent
        nested += 1
        if nested == 10:
            break
    log.debug("found {name} folder {current}".format(name=name, current=str(current_folder)))
    return current_folder.absolute()


def get_temp_file(dirname: str, filename: str):
    output_file = Path(tempfile.gettempdir(), dirname, filename)
    os.makedirs(output_file.parent)
    return output_file


def get_temp_dir(dirname: str, create: bool = True):
    temp_dir = Path(tempfile.gettempdir(), dirname)
    if create:
        temp_dir.mkdir(exist_ok=True)
    return temp_dir


def cleanup(name):
    temp_dir = get_temp_dir(name, create=False)
    if temp_dir.exists():
        shutil.rmtree(temp_dir)



def check_filename_ends_with_json(output_file_path):
    check_ends_with(".json", output_file_path)


def check_ends_with(extension, output_file_path):
    if not str(output_file_path).endswith(extension):
        expected_message = "output path must end with {ext} for example " \
                           "c:/output/reading{ext} or /home/output/reading{ext}".format(ext=extension)
        raise ValueError(expected_message)



from pathlib import Path

from marshmallow import fields, ValidationError, Schema, post_dump, post_load

from pysunspec_read.connect_options import ConnectOptions, ConnectOptionsTcp, ConnectOptionsRtu, ConnectOptionsFile
from pysunspec_read.output_options import OutputOptions


class PathField(fields.Field):
    """Field that serializes to a string
        deserializes to a Path.
    """

    def _serialize(self, value, attr, obj, **kwargs):
        if value is None:
            return None
        return str(value.as_posix())

    def _deserialize(self, value, attr, data, **kwargs):
        try:
            return Path(value)
        except ValueError as error:
            raise ValidationError("PathField must be a string") from error


class BaseSchema(Schema):
    # SKIP_VALUES = {None}

    @post_dump
    def remove_skip_values(self, data, **kwargs):
        return {
            key: value for key, value in data.items()
            # if value not in BaseSchema.SKIP_VALUES
            if value is not None
        }


class OutputOptionsSchema(BaseSchema):
    output_file_path = PathField()
    console = fields.Boolean()
    scale = fields.Boolean()
    log_reading = fields.Boolean()
    save_reading = fields.Boolean()
    omit_zero_readings = fields.Boolean()
    omit_none_readings = fields.Boolean()
    add_timestamp_to_reading = fields.Boolean()
    add_timestamp_to_filename = fields.Boolean()

    @post_load
    def make_request(self, data, **kwargs):
        return OutputOptions(**data)


class ConnectOptionsTcpSchema(BaseSchema):
    ip_address = fields.String()
    ip_port = fields.Integer()
    slave_id = fields.Integer()
    timeout = fields.Float()
    ctx = fields.String()
    max_count = fields.Integer()
    test = fields.Boolean()

    @post_load
    def make_request(self, data, **kwargs):
        return ConnectOptionsTcp(**data)


class ConnectOptionsRtuSchema(BaseSchema):
    slave_id = fields.Integer()
    name = fields.String()
    baudrate = fields.Integer()
    parity = fields.String()
    timeout = fields.Float()
    ctx = fields.Raw()
    max_count = fields.Integer()

    @post_load
    def make_request(self, data, **kwargs):
        return ConnectOptionsRtu(**data)


class ConnectOptionsFileSchema(BaseSchema):
    filename = fields.String()
    addr = fields.Integer()

    @post_load
    def make_request(self, data, **kwargs):
        return ConnectOptionsFile(**data)


class ConnectOptionsSchema(BaseSchema):
    tcp: ConnectOptionsTcp = fields.Nested(ConnectOptionsTcpSchema)
    rtu: ConnectOptionsRtu = fields.Nested(ConnectOptionsRtuSchema)
    file: ConnectOptionsFile = fields.Nested(ConnectOptionsFileSchema)

    @post_load
    def make_request(self, data, **kwargs):
        return ConnectOptions(**data)



from dataclasses import dataclass
from inspect import signature
from typing import Any

# TODO check fields on TCP and RTU for types
from sunspec2.modbus.client import SunSpecModbusClientDeviceTCP, SunSpecModbusClientDeviceRTU


@dataclass
class ConnectOptionsTcp:
    slave_id: int = signature(SunSpecModbusClientDeviceTCP).parameters["slave_id"].default
    ip_address: str = signature(SunSpecModbusClientDeviceTCP).parameters["ipaddr"].default
    ip_port: int = signature(SunSpecModbusClientDeviceTCP).parameters["ipport"].default
    timeout: float = None
    ctx: Any = None
    max_count: int = signature(SunSpecModbusClientDeviceTCP).parameters["max_count"].default
    test: bool = signature(SunSpecModbusClientDeviceTCP).parameters["test"].default


@dataclass
class ConnectOptionsRtu:
    slave_id: int
    name: str
    baudrate: int = None
    parity: str = None
    timeout: float = None
    ctx: Any = None
    max_count: int = signature(SunSpecModbusClientDeviceRTU).parameters["max_count"].default


@dataclass
class ConnectOptionsFile:
    filename: str = None
    addr: int = 40000


@dataclass
class ConnectOptions:
    file: ConnectOptionsFile = None
    tcp: ConnectOptionsTcp = None
    rtu: ConnectOptionsRtu = None



import logging.config
import os

import yaml


def setup_logging(default_path='logging.yaml', default_level=logging.INFO, env_key='LOG_CFG'):
    """Setup logging configuration

    """
    path = default_path
    value = os.getenv(env_key, None)
    if value:
        path = value
    if os.path.exists(path):
        with open(path, 'rt') as f:
            config = yaml.safe_load(f.read())
        logging.config.dictConfig(config)
    else:
        logging.basicConfig(level=default_level)



from os.path import splitext
from pathlib import Path


def add_timestamp_to_filename(output_options, reading_time):
    if output_options.add_timestamp_to_filename:
        path_without_extension = splitext(output_options.output_file_path)[0]
        epoch_seconds = int(reading_time.timestamp())
        file_date_str = reading_time.strftime("%Y-%m-%d_%H-%M-%S")
        return Path(path_without_extension + "_d_" + file_date_str + "_e_" + str(epoch_seconds) + ".json")
    else:
        return output_options.output_file_path



from numbers import Number
from typing import Any


def apply(predicates, key, value):
    result = any(p for p in predicates if p(key, value))
    return result


def is_none(key: str, value: Any) -> bool:
    return value is None


def is_zero(key: str, value: Any) -> bool:
    return isinstance(value, Number) and value == 0


def is_scale_factor(key: str, value: Any) -> bool:
    return isinstance(key, str) and str(key).casefold().endswith("_sf")



import logging
import sys

from pysunspec_read.connect_options import ConnectOptions, ConnectOptionsTcp
from pysunspec_read.output_options import OutputOptions
from pysunspec_read.read_to_output import read_with_clean

logging.basicConfig(level=logging.INFO)

logger = logging.getLogger(__name__)
logger.info("****************************")
logger.info("*         Starting         *")
logger.info("****************************")


def demo_read_tcp(ip_address: str):
    print("Starting demo reading from {}".format(ip_address))
    logger.info("Checking logging")
    if len(sys.argv) > 1:
        logger.info("using IP address from command line: {}".format(sys.argv[1]))
        ip_address = sys.argv[1]

    read_with_clean(connect_options=ConnectOptions(tcp=ConnectOptionsTcp(ip_address=ip_address)),
                    output_options=OutputOptions(save_reading=False, log_reading=True,
                                                 add_timestamp_to_reading=True))


def demo_read_tcp_to_file(ip_address: str, file_path: str):
    print("Starting demo reading from {} to file {}".format(ip_address, file_path))
    logger.info("Checking logging")
    if len(sys.argv) > 2:
        logger.info("using IP address and file {} from command line: {}".format(sys.argv[1], sys.argv[2]))
        ip_address = sys.argv[1]
        file_path = sys.argv[2]

    read_with_clean(connect_options=ConnectOptions(tcp=ConnectOptionsTcp(ip_address=ip_address)),
                    output_options=OutputOptions(save_reading=True, output_file_path=file_path, log_reading=True,
                                                 add_timestamp_to_reading=True))


def usage():
    print("For a demo reading, call instead to demo_ip.py")


if __name__ == '__main__':
    usage()



from setuptools import find_packages
from setuptools import setup
from os.path import splitext
from os.path import basename
from glob import glob

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="pysunspec_read",
    version="0.20.4",
    package_dir={"": "src"},
    # packages=[""],
    packages=find_packages("src"),
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    url="https://bitbucket.org/clockworkcodeteam/pysunspec-read/src/master/",
    license="GNU General Public License V3 or later",
    author="dilbertau99",
    author_email="dilbertau99@gmail.com",
    description="read from solar inverters that implement the SunSpec interface, to simpler datastructure",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="solar monitoring inverter communications reader",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
        "Development Status :: 1 - Planning",
        "Intended Audience :: End Users/Desktop",
        "Natural Language :: English",
        "Topic :: Utilities",
        "Topic :: Communications",
    ],
    install_requires=["pysunspec2", "pyserial", "pyyaml"],
    tests_require= ["deepdiff"]
)



from typing import Callable, Any


def clean(obj: Any, reject: Callable[[str, Any], bool]) -> Any:
    """
    Recursively remove all None values from dictionaries and lists, and returns
    the result as a new dictionary or list.
    """
    if isinstance(obj, list):
        return clean_list(obj, reject)
    elif isinstance(obj, dict):
        return clean_dict(obj, reject)
    else:
        return obj


def clean_dict(dict_to_clean: dict, reject: Callable[[str, Any], bool]) -> dict:
    cleaned_dict = {key: clean(value, reject) for key, value in dict_to_clean.items() if not reject(key, value)}
    return cleaned_dict


def clean_list(list_to_clean: list, reject: Callable[[str, Any], bool]) -> list:
    cleaned_list = [clean(list_item, reject) for list_item in list_to_clean if not reject(None, list_item)]
    return cleaned_list



# This file is part of PySunsSpec-Read
#
# PySunSpec-Read is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     PySunSpec-Read is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with PySunSpec-Read.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2020 David Smith applies to this and each file in this project

import json
import logging
from datetime import datetime

from .clean_dict import clean
from .connect_options import ConnectOptions
from .dates import add_timestamp_to_filename
from .device_reader import read
from .files import check_filename_ends_with_json
from .output_filter_predicates import is_zero, is_scale_factor, is_none, apply
from .output_options import OutputOptions

logger = logging.getLogger(__name__)


def read_with_clean(connect_options: ConnectOptions, output_options: OutputOptions) -> str:
    """Reads data from a SunSpec device and outputs it as json, for example from a solar inverter.

    :return: json string, and if output_options.save_reading is True,
        will save the output to a file specified by output_options.output_file_path
    """
    if output_options.save_reading:
        assert output_options.output_file_path
        check_filename_ends_with_json(output_options.output_file_path)
    reading_time = datetime.now()
    reading = read_from_device(connect_options, output_options, reading_time)
    cleaned_reading = clean_reading(output_options, reading, reading_time)
    reading_json = write_as_json(cleaned_reading, output_options, reading_time)
    return reading_json


def read_from_device(connect_options: ConnectOptions, output_options: OutputOptions, reading_time: datetime) -> dict:
    """Intention is that you would call read_with_clean"""
    logger.info("Reading time: " + reading_time.strftime("%Y-%m-%d %H:%M:%S"))
    with read(connect_options=connect_options) as device:
        reading = device.get_dict(computed=not output_options.scale)
    return reading


def write_as_json(cleaned_reading: dict, output_options: OutputOptions, reading_time: datetime) -> str:
    """Intention is that you would call read_with_clean"""
    reading_json = json.dumps(cleaned_reading, sort_keys=False, indent=3)
    if output_options.log_reading:
        logger.info("cleaned reading: " + reading_json)
    if output_options.save_reading:
        with open(add_timestamp_to_filename(output_options, reading_time), "w") as out:
            logger.info("writing reading to {}".format(out.name))
            out.write(reading_json)
    return reading_json


# This would probably be nicer with an interface for a predicate function
def curry_predicates(preds):
    def apply_predicates(key, value):
        return apply(preds, key, value)

    return apply_predicates


def clean_reading(output_options: OutputOptions, reading: dict, reading_time: datetime) -> dict:
    """Intention is that you would call read_with_clean"""

    list_of_predicate_funcs = create_predicates(output_options)
    predicate_list_applier = curry_predicates(list_of_predicate_funcs)
    cleaned_reading = clean(reading, predicate_list_applier)
    if output_options.add_timestamp_to_reading:
        iso_format = reading_time.isoformat()
        logger.info("adding reading_date to output ({})".format(iso_format))
        cleaned_reading["reading_date"] = iso_format
    return cleaned_reading


def create_predicates(output_options: OutputOptions):
    predicates = []
    if output_options.omit_zero_readings:
        predicates.append(is_zero)

    if output_options.omit_none_readings:
        predicates.append(is_none)

    if not output_options.scale:
        predicates.append(is_scale_factor)

    return predicates
